# Copyright  2017  Department of Biomedical Informatics, University of Utah
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# from quicksect import IntervalTree

from .WildCardFunctions import WildCardFunctions
from .IOUtils import IOUtils, Rule, Span
from .ReplicationFunctionsLambda import ReplicationFunctions
from .ReplicationFunctionsLambda import processReplicationCommon
import os
import logging
import logging.config

from quicksectx.quicksectx cimport IntervalTree, Interval



cpdef class CFastCNER:
    END = ('<END>')

    def __init__(self, rule_str='', max_repeat=50, enable_logger=False):
        self.span_compare_method = 'width'
        self.support_replication = False
        self.offset = 0
        self.max_repeat = max_repeat
        self.overlap_checkers = dict()
        self.wildcard_funcs = WildCardFunctions(self.processRules).wildcard_funcs
        self.replication_funcs = ReplicationFunctions(self.processRules, self.max_repeat).replication_funcs
        self.rule_str = rule_str
        self.rule_map = dict()
        self.scores = dict()
        self.full_definition = False
        self.rule_store = {}
        self.initiate(rule_str)
        if enable_logger:
            self.logger == logging.getLogger(__name__)
        else:
            self.logger = None

    def initiate(self, rule_str):
        self.rule_str = rule_str
        self.rule_map.clear()
        self.rule_store.clear()
        self.scores.clear()
        io_utils = IOUtils(rule_str)
        self.full_definition = io_utils.full_definition
        self.rule_store = io_utils.rule_cells
        self.constructRuleMap(self.rule_store)
        pass

    def constructRuleMap(self, rule_store):
        # reset the variables in case reuse this function to re-initiate from rule_store
        self.rule_map.clear()
        self.rule_store = rule_store
        self.scores.clear()

        for id, rule in rule_store.items():
            if '[' in rule.rule:
                rules = self.expandSquareBracket(rule)
                for subrule in rules:
                    self.addRule(subrule)
            else:
                self.addRule(rule)
        pass

    def getType(self, span):
        return self.rule_store[span.rule_id].type

    # expand rules with square brackets
    def expandSquareBracket(self, rule):
        OUT = 0
        IN = 1
        expanded_rules = []
        rule_builder = []
        rule_builder.append([])
        rule_str = rule.rule
        # status: 0 for OUT, 1 for in
        status = OUT
        pre_char = ' '
        next_ch = ' '
        branches = []
        for i in range(0, len(rule_str)):
            ch = rule_str[i]
            if i > 0:
                pre_char = rule_str[i - 1]
            if i < len(rule_str) - 1:
                next_ch = rule_str[i + 1]
            else:
                next_ch = ' '
            if status == OUT and (ch != '[' or pre_char == '\\'):
                if ch == '\\' and (next_ch == '[' or next_ch == ']'):
                    pre_char = ch
                    continue
                for j in range(0, len(rule_builder)):
                    rule_builder[j].append(ch)
            elif status == OUT and ch == '[' and pre_char != '\\':
                status = IN
                branches = [[]]
                continue
            elif status == IN and (ch != ']' or (ch == ']' and pre_char == '\\')):
                if ch == '|':
                    branches.append([])
                elif ch == '\\' and (next_ch == '[' or next_ch == ']'):
                    pre_char = ch
                    continue
                else:
                    branches[-1].append(ch)
            elif status == IN and ch == ']' and pre_char != '\\':
                status = OUT
                previous_size = len(rule_builder)
                if ch == '\\' and (next_ch == '[' or next_ch == ']'):
                    pre_char = ch
                    continue
                for j in range(0, previous_size):
                    sb = list(rule_builder[j])
                    rule_builder[j].extend(branches[0])
                    for k in range(1, len(branches)):
                        rule_builder.append(list(sb))
                        rule_builder[-1].extend(branches[k])
        cleanSet = set()
        for sub_rule in rule_builder:
            rule_str = ''.join(sub_rule)
            if rule_str not in cleanSet:
                new_rule = Rule(rule.id, rule_str, rule.rule_name, rule.score, rule.type)
                expanded_rules.append(new_rule)
                cleanSet.add(rule_str)
        return expanded_rules

    def addRule(self, rule):
        determinant = rule.rule_name
        rule_map1 = self.rule_map
        rule_map2 = dict()
        length = len(rule.rule)
        rule_str = rule.rule
        if '+' in rule_str:
            self.support_replication = True
        i = 0
        while i < length and rule_map1 != None and rule_str[i] in rule_map1:
            rule_map1 = rule_map1.get(rule_str[i])
            i += 1
        # if the rule has been included
        if i == length and CFastCNER.END in rule_map1 and rule_map1[CFastCNER.END] == determinant:
            return False
        # 		start with the determinant, construct the last descendant dict
        if i == length:
            if CFastCNER.END in rule_map1:
                rule_map1[CFastCNER.END][determinant] = rule.id
            else:
                rule_map2[determinant] = rule.id
                rule_map1[CFastCNER.END] = rule_map2.copy()
            self.scores[rule.id] = rule.score
            return True
        else:
            rule_map2[determinant] = rule.id
            rule_map2[CFastCNER.END] = rule_map2.copy()
            rule_map2.pop(determinant, None)

            for j in range(length - 1, i, -1):
                rule_mapt = rule_map2.copy()
                rule_map2.clear()
                rule_map2[rule_str[j]] = rule_mapt
        self.scores[rule.id] = rule.score
        rule_map1[rule_str[i]] = rule_map2.copy()
        return True

    def processString(self, text, offset=0):
        self.offset = offset
        self.overlap_checkers.clear()
        matches = dict()
        for i in range(0, len(text)):
            pre_char = text[i - 1] if i > 0 else ' '
            self.processRules(text, self.rule_map, i, 0, i, matches, pre_char, False, ' ')
        self.removePseudo(matches)
        return matches

    def processRules(self, text, rule_map, match_begin, match_end, current_position,
                     matches, pre_char, wildcard_support, pre_key):
        length = len(text)
        if current_position < length:
            this_char = text[current_position]

            if '\\' in rule_map:
                self.processWildCards(text, rule_map['\\'], match_begin, match_end, current_position, matches, pre_char,
                                      True, '\\')
            if '(' in rule_map and pre_key != '\\':
                self.processRules(text, rule_map['('], current_position, match_end, current_position, matches, pre_char,
                                  False, '(')
            if ')' in rule_map and pre_key != '\\':
                self.processRules(text, rule_map[')'], match_begin, current_position, current_position, matches,
                                  pre_char, False, ')')

            if CFastCNER.END in rule_map:
                self.addDeterminants(text, rule_map, matches, match_begin, match_end, current_position)
            if this_char in rule_map and this_char != ')' and this_char != '(':
                self.processRules(text, rule_map[this_char], match_begin, match_end, current_position + 1, matches,
                                  this_char, False, this_char)

            if self.support_replication and '+' in rule_map:
                self.processRules(text, rule_map['+'], match_begin, match_end, current_position, matches, this_char,
                                  False, '+')
                self.processReplication(text, rule_map['+'], match_begin, match_end, current_position, matches,
                                        this_char, wildcard_support, pre_key)
        elif current_position == length and CFastCNER.END in rule_map:
            if match_end == 0:
                self.addDeterminants(text, rule_map, matches, match_begin, current_position, current_position)
            else:
                self.addDeterminants(text, rule_map, matches, match_begin, match_end, current_position)
        elif current_position == length and '\\' in rule_map and 'e' in rule_map['\\']:
            deter_rule = rule_map['\\']['e']
            if match_end == 0:
                self.addDeterminants(text, deter_rule, matches, match_begin, current_position, current_position)
            else:
                self.addDeterminants(text, deter_rule, matches, match_begin, match_end, current_position)
        elif current_position == length and ')' in rule_map:
            deter_rule = rule_map[')']
            if CFastCNER.END in deter_rule:
                self.addDeterminants(text, deter_rule, matches, match_begin, current_position, current_position)
            elif '\\' in deter_rule and 'e' in deter_rule['\\']:
                self.processRules(text, deter_rule['\\'], match_begin, current_position, current_position, matches,
                                  pre_char, False, ' ')
        elif current_position == length and '+' in rule_map:
            self.processRules(text, rule_map['+'], match_begin, match_end, current_position, matches, pre_char,
                              wildcard_support, pre_key)
        pass

    def processReplication(self, text, rule_map, match_begin, match_end, current_position, matches,
                           pre_char, wildcard_support, pre_key):
        this_char = text[current_position]
        if wildcard_support:
            if pre_key in self.replication_funcs:
                self.replication_funcs[pre_key](text, rule_map, match_begin, match_end, current_position, matches,
                                                this_char, pre_char)
            else:
                logging.info('"\\' + pre_key + '+" is not a eligible syntax.')
        else:
            processReplicationCommon(lambda char: char == pre_key, self.processRules, text, rule_map, match_begin,
                                     match_end, current_position,
                                     matches,
                                     this_char, pre_char)
        pass

    def processWildCards(self, text, rule_map, match_begin, match_end, current_position, matches,
                         pre_char, wildcard_support, pre_key):
        this_char = text[current_position]
        for rule_char in rule_map.keys():
            if rule_char in self.wildcard_funcs:
                self.wildcard_funcs[rule_char](text, rule_map, match_begin, match_end, current_position, matches,
                                               this_char)
            else:
                logging.info('"\\' + rule_char + '" is not a eligible syntax.')
        pass

    def addDeterminants(self, text, deter_rule, matches, match_begin, match_end, current_position):
        deter_rule = deter_rule[CFastCNER.END]
        end = current_position if match_end == 0 else match_end
        current_span = Span(match_begin + self.offset, end + self.offset,
                            text[match_begin:end])
        current_spans_list = []
        overlap_checkers = self.overlap_checkers
        for key in deter_rule.keys():
            rule_id = deter_rule[key]
            if self.logger is not None:
                self.logger.debug(
                    'try add matched rule ({}-{})\t{}'.format(match_begin, match_end, str(self.rule_store[rule_id])))
            current_span.rule_id = rule_id
            if key in matches:
                current_spans_list = matches[key]
                overlap_checker = overlap_checkers[key]
                overlapped_pos = overlap_checker.search(current_span.begin, current_span.end)
                if len(overlapped_pos) > 0:
                    pos = overlapped_pos.pop().data
                    overlapped_span = current_spans_list[pos]
                    if not self.compareSpan(current_span, overlapped_span):
                        continue
                    current_spans_list[pos] = current_span
                    overlap_checker.remove(Interval(current_span.begin, current_span.end))
                    overlap_checker.add(current_span.begin, current_span.end, pos)
                else:
                    overlap_checker.add(current_span.begin, current_span.end, len(current_spans_list))
                    current_spans_list.append(current_span)
            else:
                matches[key] = current_spans_list
                overlap_checker = IntervalTree()
                # quickset's search will include both lower and upper bounds, so minus one from the end.
                overlap_checker.add(current_span.begin, current_span.end - 1, len(current_spans_list))
                current_spans_list.append(current_span)
                overlap_checkers[key] = overlap_checker

        pass

    def compareSpan(self, a, b):
        if self.span_compare_method == 'score':
            return a.score < 0 or a.score > b.score
        elif self.span_compare_method == 'scorewidth':
            return a.score < 0 or a.score > b.score or (a.score == b.score and a.width > b.width)
        elif self.span_compare_method == 'widthscore':
            return a.width > b.width or (a.width == b.width and a.score > b.score)
        else:
            return a.width > b.width

    def removePseudo(self, matches):
        for key, span_list in matches.items():
            cleaned_list = [x for x in span_list if self.rule_store[x.rule_id].type == 'ACTUAL']
            matches[key] = cleaned_list
        pass



# Copyright  2017  Department of Biomedical Informatics, University of Utah
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import csv
import logging
from decimal import Decimal
from typing import Union, List, Iterator, Generator


class IOUtils:

    def __init__(self, rules: Union[str, List]):
        self.rule_cells = {}
        self.initiations = {}
        self.full_definition = False
        if isinstance(rules, str):
            if rules.lower().endswith('csv'):
                self.read(rules, ',')
            elif rules.lower().endswith('tsv'):
                self.read(rules, '\t')
            else:
                self.readString(rules, '\t')
        elif isinstance(rules, List):
            if len(rules) > 0:
                if isinstance(rules[0], List):
                    self.parse_iterator(rules)
                elif isinstance(rules[0], str):
                    self.parse_iterator([line.split('\t') for line in rules])
            pass
        else:
            raise ValueError(
                "Rules can either be a List or a string. The input type: '{}' is not eligible.".format(type(rules)))
        pass

    def read(self, file_name, delimiter):
        with open(file_name, newline='') as csvfile:
            self.parse(csvfile, delimiter)
        pass

    def readString(self, input, delimiter):
        self.parse(input.splitlines(), delimiter)

    def parse(self, input, delimiter):
        spamreader = csv.reader(input, delimiter=delimiter)
        self.parse_iterator(spamreader)
        pass

    def parse_iterator(self, iterator: Union[Iterator, List]):
        row_num = 0
        for row in iterator:
            row_num += 1
            if len(row) < 2:
                continue
            if row[0].startswith('#'):
                continue
            if row[0].startswith('@'):
                if row[0][1].isupper():
                    self.initiations[row_num] = row
            elif len(row) >= 3:
                self.rule_cells[row_num] = self.buildRule(row_num, row)
            else:
                logging.info('Incorrect formated rule: ' + str(row))
        pass

    def buildRule(self, row_num, row):
        rule_type = 'ACTUAL'
        if len(row) == 4:
            rule_type = row[3].strip()
            self.full_definition = True
        return Rule(row_num, row[0], row[2].strip(), float(row[1].strip()), rule_type)


class Rule:
    def __init__(self, id, rule, rule_name, score=1.0, type='ACTUAL'):
        self.id = id
        self.rule = rule
        self.score = score
        self.rule_name = rule_name
        self.type = type

    def copy(self):
        return Rule(self.id, self.rule, self.rule_name, self.score, self.type)

    def __str__(self):
        return "Rule {0}:\n\trule content:\t{1}\n\trule score:\t{2}\n\trule name:\t{3}\n\trule type:\t{4}\n".format(
            self.id, self.rule, self.score, self.rule_name, self.type)

    def __repr__(self):
        return self.__str__()


class Span:
    def __init__(self, begin=-1, end=-1, text='', rule_id=-1, score=1.0):
        self.begin = begin
        self.end = end
        self.width = end - begin
        self.text = text
        self.rule_id = rule_id
        self.score = score

    def __str__(self):
        return "Span: \n\tbegin:\t{0}\n\tend:\t{1}\n\twidth:\t{2}\n\ttext:\t{3}\n\trule_id:\t{4}\n\tscore:\t{5}".format(
            self.begin, self.end, self.width, self.text, self.rule_id, self.score
        )



# Copyright  2017  Department of Biomedical Informatics, University of Utah
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import string


class WildCardFunctions:
    def __init__(self, processRulesFunc):
        self.wildcard_funcs = dict()
        self.initWildCardFuncs(self.wildcard_funcs)
        self.processRules = processRulesFunc
        pass

    def initWildCardFuncs(self, wildcard_funs):
        wildcard_funs['s'] = self.processWildCard_s
        wildcard_funs['n'] = self.processWildCard_n
        wildcard_funs['('] = self.processWildCard_openParan
        wildcard_funs[')'] = self.processWildCard_closeParan
        wildcard_funs['d'] = self.processWildCard_d
        wildcard_funs['C'] = self.processWildCard_C
        wildcard_funs['c'] = self.processWildCard_c
        wildcard_funs['p'] = self.processWildCard_p
        wildcard_funs['+'] = self.processWildCard_plus
        wildcard_funs['\\'] = self.processWildCard_backSlash
        wildcard_funs['b'] = self.processWildCard_b
        wildcard_funs['a'] = self.processWildCard_a
        wildcard_funs['u'] = self.processWildCard_u
        wildcard_funs['w'] = self.processWildCard_w
        wildcard_funs['e'] = self.processWildCard_e
        pass

    def processWildCard_s(self, text, rule_map, match_begin, match_end, current_position, matches,
                          this_char):
        if this_char == ' ' or this_char == '\t' or ord(this_char) == 160:
            self.processRules(text, rule_map['s'], match_begin, match_end, current_position + 1, matches,
                              this_char, True, 's')
        pass

    def processWildCard_n(self, text, rule_map, match_begin, match_end, current_position, matches,
                          this_char):
        if this_char == '\n' or this_char == '\r':
            self.processRules(text, rule_map['n'], match_begin, match_end, current_position + 1, matches,
                              this_char, True, 'n')
        pass

    def processWildCard_openParan(self, text, rule_map, match_begin, match_end, current_position, matches,
                                  this_char):
        if this_char == '(':
            self.processRules(text, rule_map['('], match_begin, match_end, current_position + 1, matches,
                              this_char, True, '(')
        pass

    def processWildCard_closeParan(self, text, rule_map, match_begin, match_end, current_position, matches,
                                   this_char):
        if this_char == ')':
            self.processRules(text, rule_map[')'], match_begin, match_end, current_position + 1, matches,
                              this_char, True, ')')
        pass

    def processWildCard_d(self, text, rule_map, match_begin, match_end, current_position, matches,
                          this_char):
        if this_char.isdigit():
            self.processRules(text, rule_map['d'], match_begin, match_end, current_position + 1, matches,
                              this_char, True, 'd')
        pass

    def processWildCard_C(self, text, rule_map, match_begin, match_end, current_position, matches,
                          this_char):
        if this_char.isupper():
            self.processRules(text, rule_map['C'], match_begin, match_end, current_position + 1, matches,
                              this_char, True, 'C')
        pass

    def processWildCard_c(self, text, rule_map, match_begin, match_end, current_position, matches,
                          this_char):
        if this_char.islower():
            self.processRules(text, rule_map['c'], match_begin, match_end, current_position + 1, matches,
                              this_char, True, 'c')
        pass

    def processWildCard_p(self, text, rule_map, match_begin, match_end, current_position, matches,
                          this_char):
        if this_char in string.punctuation:
            self.processRules(text, rule_map['p'], match_begin, match_end, current_position + 1, matches,
                              this_char, True, 'p')
        pass

    def processWildCard_plus(self, text, rule_map, match_begin, match_end, current_position, matches,
                             this_char):
        if this_char.islower():
            self.processRules(text, rule_map['c'], match_begin, match_end, current_position + 1, matches,
                              this_char, True, 'c')
        pass

    def processWildCard_backSlash(self, text, rule_map, match_begin, match_end, current_position, matches,
                                  this_char):
        if this_char == '\\':
            self.processRules(text, rule_map['\\'], match_begin, match_end, current_position + 1, matches,
                              this_char, True, '\\')

    def processWildCard_a(self, text, rule_map, match_begin, match_end, current_position, matches,
                          this_char):
        if not this_char.isspace():
            self.processRules(text, rule_map['a'], match_begin, match_end, current_position + 1, matches,
                              this_char, True, 'a')
            pass

    def processWildCard_u(self, text, rule_map, match_begin, match_end, current_position, matches,
                          this_char):
        if this_char > '~' and ord(this_char) != 160:
            self.processRules(text, rule_map['u'], match_begin, match_end, current_position + 1, matches,
                              this_char, True, 'u')
            pass

    def processWildCard_w(self, text, rule_map, match_begin, match_end, current_position, matches,
                          this_char):
        if this_char > '~' or this_char.isspace():
            self.processRules(text, rule_map['w'], match_begin, match_end, current_position + 1, matches,
                              this_char, True, 'w')
            pass

    def processWildCard_b(self, text, rule_map, match_begin, match_end, current_position, matches,
                          this_char):
        if current_position == 0:
            self.processRules(text, rule_map['b'], match_begin, match_end, current_position, matches,
                              this_char, True, 'b')
            pass

    def processWildCard_e(self, text, rule_map, match_begin, match_end, current_position, matches,
                          this_char):
        if current_position == len(text):
            self.processRules(text, rule_map['e'], match_begin, match_end, current_position, matches,
                              this_char, True, 'e')
            pass